from gi.repository import Gtk, GObject, Gdk, Pango

from pychess.Utils.IconLoader import load_icon
from pychess.widgets.InfoPanel import Panel

TYPE_PERSONAL, TYPE_CHANNEL, TYPE_GUEST, TYPE_ADMIN, TYPE_COMP, TYPE_BLINDFOLD = range(
    6
)

add_icon = load_icon(16, "gtk-add", "list-add")
remove_icon = load_icon(16, "gtk-remove", "list-remove")


def cmp(x, y):
    return (x > y) - (x < y)


class TextImageTree(Gtk.TreeView):
    """:Description: Defines a tree with two columns.
    The first one has text. The second one a clickable stock_icon
    """

    __gsignals__ = {
        "activated": (GObject.SignalFlags.RUN_FIRST, None, (str, str, int)),
        "selected": (GObject.SignalFlags.RUN_FIRST, None, (str, int)),
    }

    def __init__(self, icon):
        GObject.GObject.__init__(self)
        self.id2iter = {}

        pm = Gtk.ListStore(str, str, int, str)
        self.sort_model = Gtk.TreeModelSort(model=pm)
        self.set_model(self.sort_model)
        self.idSet = set()

        self.set_headers_visible(False)
        self.set_tooltip_column(3)
        self.set_search_column(1)
        self.sort_model.set_sort_column_id(1, Gtk.SortType.ASCENDING)
        self.sort_model.set_sort_func(1, self.compareFunction, 1)

        # First column
        crp = Gtk.CellRendererPixbuf()
        crp.props.pixbuf = icon
        self.rightcol = Gtk.TreeViewColumn("", crp)
        self.append_column(self.rightcol)

        # Second column
        crt = Gtk.CellRendererText()
        crt.props.ellipsize = Pango.EllipsizeMode.END
        self.leftcol = Gtk.TreeViewColumn("", crt, text=1)
        self.leftcol.set_expand(True)
        self.append_column(self.leftcol)

        # Mouse
        self.pressed = None
        self.stdcursor = Gdk.Cursor.new(Gdk.CursorType.LEFT_PTR)
        self.linkcursor = Gdk.Cursor.new(Gdk.CursorType.HAND2)
        self.connect("button_press_event", self.button_press)
        self.connect("button_release_event", self.button_release)
        self.connect("motion_notify_event", self.motion_notify)
        self.connect("leave_notify_event", self.leave_notify)

        # Selection
        self.get_selection().connect("changed", self.selection_changed)

    def addRow(self, grp_id, text, grp_type):
        """:Description: Takes a player or a channel identified by grp_id and adds
        them to the correct group defined by grp_type
        :return: None
        """
        if grp_id in self.id2iter:
            return
        model = self.sort_model.get_model()
        m_iter = model.append(
            [grp_id, text, grp_type, GObject.markup_escape_text(text)]
        )
        self.id2iter[grp_id] = m_iter
        self.idSet.add(grp_id)

    def removeRow(self, grp_id):
        """:Description: Takes a player or channel identified by grp_id and removes them from
        the data model.
        :return: None
        """
        try:
            m_iter = self.id2iter[grp_id]
        except KeyError:
            return
        model = self.sort_model.get_model()
        model.remove(m_iter)
        del self.id2iter[grp_id]
        self.idSet.remove(grp_id)

    def selectRow(self, grp_id):
        """:Description: Takes a grp_id and finds the row associated with this id then
        sets this row to be the focus ie selected

        :returns: None
        """
        m_iter = self.id2iter[grp_id]
        m_iter = self.sort_model.convert_child_iter_to_iter(m_iter)[1]
        sel = self.get_selection()
        sel.select_iter(m_iter)

    def __contains__(self, grp_id):
        """:Description: Checks to see if a grp_id in a member of the id set
        :returns: boolean
        """
        return grp_id in self.idSet

    def button_press(self, widget, event):
        path_col_pos = self.get_path_at_pos(int(event.x), int(event.y))
        if path_col_pos and path_col_pos[1] == self.rightcol:
            self.pressed = path_col_pos[0]

    def button_release(self, widget, event):
        path_col_pos = self.get_path_at_pos(int(event.x), int(event.y))
        if path_col_pos and path_col_pos[1] == self.rightcol:
            if self.pressed == path_col_pos[0]:
                model = self.sort_model
                m_iter = model.get_iter(self.pressed)
                grp_id = model.get_value(m_iter, 0)
                text = model.get_value(m_iter, 1)
                grp_type = model.get_value(m_iter, 2)
                self.emit("activated", grp_id, text, grp_type)
        self.pressed = None

    def motion_notify(self, widget, event):
        path_col_pos = self.get_path_at_pos(int(event.x), int(event.y))
        if path_col_pos and path_col_pos[1] == self.rightcol:
            self.get_window().set_cursor(self.linkcursor)
        else:
            self.get_window().set_cursor(self.stdcursor)

    def leave_notify(self, widget, event):
        self.get_window().set_cursor(self.stdcursor)

    def selection_changed(self, selection):
        model, m_iter = selection.get_selected()
        if m_iter:
            grp_id = model.get_value(m_iter, 0)
            grp_type = model.get_value(m_iter, 2)
            self.emit("selected", grp_id, grp_type)

    def compareFunction(self, treemodel, iter0, iter1, column):
        val0 = treemodel.get_value(iter0, column).split(":")[0]
        val1 = treemodel.get_value(iter1, column).split(":")[0]
        if val0.isdigit() and val1.isdigit():
            return cmp(int(val0), int(val1))
        return cmp(val0, val1)


class ChannelsPanel(Gtk.ScrolledWindow, Panel):
    __gsignals__ = {
        "conversationAdded": (GObject.SignalFlags.RUN_FIRST, None, (str, str, int)),
        "conversationRemoved": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        "conversationSelected": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection

        self.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        vbox = Gtk.VBox()
        self.add_with_viewport(vbox)
        self.get_child().set_shadow_type(Gtk.ShadowType.NONE)

        self.joinedList = TextImageTree(remove_icon)
        self.joinedList.connect("activated", self.onRemove)
        self.joinedList.connect("selected", self.onSelect)
        vbox.pack_start(self.joinedList, True, True, 0)

        vbox.pack_start(Gtk.Separator.new(0), False, False, 2)
        expander = Gtk.Expander.new(_("Friends"))
        vbox.pack_start(expander, False, True, 0)
        self.friendsList = TextImageTree(add_icon)
        self.friendsList.connect("activated", self.onAdd)
        self.friendsList.fixed_height_mode = True
        connection.cm.connect("privateMessage", self.onPersonMessage)
        connection.cm.connect("channelsListed", self.onChannelsListed)
        vbox.pack_start(Gtk.Separator.new(0), False, False, 2)
        expander.add(self.friendsList)
        self.channels = {}

        expander = Gtk.Expander.new(_("Admin"))
        vbox.pack_start(expander, False, True, 0)
        self.adminList = TextImageTree(add_icon)
        self.adminList.connect("activated", self.onAdd)
        self.adminList.fixed_height_mode = True
        connection.cm.connect("privateMessage", self.onPersonMessage)
        connection.cm.connect("channelsListed", self.onChannelsListed)
        vbox.pack_start(Gtk.Separator.new(0), False, False, 2)
        expander.add(self.adminList)

        expander = Gtk.Expander.new(_("More channels"))
        vbox.pack_start(expander, False, True, 0)
        self.channelsList = TextImageTree(add_icon)
        self.channelsList.connect("activated", self.onAdd)
        self.channelsList.fixed_height_mode = True
        vbox.pack_start(Gtk.Separator.new(0), False, False, 2)
        expander.add(self.channelsList)

        expander = Gtk.Expander.new(_("More players"))
        vbox.pack_start(expander, False, True, 0)
        self.playersList = TextImageTree(add_icon)
        self.playersList.connect("activated", self.onAdd)
        self.playersList.fixed_height_mode = True
        connection.cm.connect("privateMessage", self.onPersonMessage)
        connection.cm.connect("channelsListed", self.onChannelsListed)
        vbox.pack_start(Gtk.Separator.new(0), False, False, 2)
        expander.add(self.playersList)

        expander = Gtk.Expander.new(_("Computers"))
        vbox.pack_start(expander, False, True, 0)
        self.compList = TextImageTree(add_icon)
        self.compList.connect("activated", self.onAdd)
        self.compList.fixed_height_mode = True
        connection.cm.connect("privateMessage", self.onPersonMessage)
        connection.cm.connect("channelsListed", self.onChannelsListed)
        vbox.pack_start(Gtk.Separator.new(0), False, False, 2)
        expander.add(self.compList)

        expander = Gtk.Expander.new(_("BlindFold"))
        vbox.pack_start(expander, False, True, 0)
        self.blindList = TextImageTree(add_icon)
        self.blindList.connect("activated", self.onAdd)
        self.blindList.fixed_height_mode = True
        connection.cm.connect("privateMessage", self.onPersonMessage)
        connection.cm.connect("channelsListed", self.onChannelsListed)
        vbox.pack_start(Gtk.Separator.new(0), False, False, 2)
        expander.add(self.blindList)

        expander = Gtk.Expander.new(_("Guests"))
        vbox.pack_start(expander, False, True, 0)
        self.guestList = TextImageTree(add_icon)
        self.guestList.connect("activated", self.onAdd)
        self.guestList.fixed_height_mode = True
        connection.cm.connect("privateMessage", self.onPersonMessage)
        connection.cm.connect("channelsListed", self.onChannelsListed)
        vbox.pack_start(Gtk.Separator.new(0), False, False, 2)
        expander.add(self.guestList)

        self.channels = {}
        self.highlighted = {}

    def change_fg_colour(self, lc, cell, model, m_iter, data):
        """
        :Description: Changes the foreground colour of a cell

        :param lc: :class:`Gtk.TreeViewColumn` The column we are interested in
        :param cell: :class:`Gtk.CellRenderer` The cell we want to change
        :param model: :class:`Gtk.TreeModel`
        :param iter: :class:`Gtk.TreeIter`
        :param data: :py:class:`dict` (key=int,value=bool) value is true if channel already highlighted
        :return: None
        """

        for chan in data:
            if model[m_iter][0] == chan:
                if data[chan]:
                    cell.set_property("foreground_rgba", Gdk.RGBA(0.9, 0.2, 0.2, 1))
                else:
                    cell.set_property("foreground_rgba", Gdk.RGBA(0, 0, 0, 1))

    def channel_Highlight(self, a, channel, grp_type, b):
        """
        :Description: Highlights a channel ( that is **not** in focus ) that has received an update and
        changes it's foreground colour to represent change in contents

        :param a: not used
        :param channel: **(str)** The channel the message is intended for
        :param grp_type: either TYPE_CHANNEL or TYPE_PERSONAL
        :param b:  not used
        :return: None
        """
        j_list = self.joinedList
        leftcol = j_list.leftcol  # treeViewColumn

        cur_iter = j_list.get_selection().get_selected()[1]  # Selected iter
        if grp_type == TYPE_PERSONAL:
            channel = "person" + channel.lower()
        tmp_iter = j_list.id2iter[channel]
        tmp_iter = j_list.sort_model.convert_child_iter_to_iter(tmp_iter)[
            1
        ]  # channel iter
        j_list.get_selection().select_iter(tmp_iter)
        cell = leftcol.get_cells()[0]
        j_list.get_selection().select_iter(cur_iter)
        self.highlighted[channel] = True
        if cur_iter != tmp_iter:
            # iter = tmp_iter
            leftcol.set_cell_data_func(
                cell, self.change_fg_colour, func_data=self.highlighted
            )

    def start(self):
        self.channels = self.connection.cm.getChannels()
        if self.channels:
            self._addChannels(self.channels)
        for player in list(self.connection.players.values()):
            grp_id = self.compileId(player.name, TYPE_PERSONAL)
            if str(player.name) in self.connection.notify_users:
                self.friendsList.addRow(
                    grp_id, player.name + player.display_titles(), TYPE_PERSONAL
                )
            elif player.online and ("(B)" in player.display_titles()):
                self.blindList.addRow(
                    grp_id, player.name + player.display_titles(), TYPE_BLINDFOLD
                )
            elif player.online and ("(C)" in player.display_titles()):
                self.compList.addRow(
                    grp_id, player.name + player.display_titles(), TYPE_COMP
                )
            elif player.online and ("Guest" in str(player.name)):
                self.guestList.addRow(
                    grp_id, player.name + player.display_titles(), TYPE_GUEST
                )
            elif player.online:
                self.playersList.addRow(
                    grp_id, player.name + player.display_titles(), TYPE_PERSONAL
                )

        def addPlayer(players, new_players):
            for player in new_players:
                # print("Player : %s : %s" % (str(player.name),player.display_titles()))
                if str(player.name) in self.connection.notify_users:
                    self.friendsList.addRow(
                        self.compileId(player.name, TYPE_PERSONAL),
                        player.name + player.display_titles(),
                        TYPE_PERSONAL,
                    )
                elif "(C)" in str(player.display_titles()):
                    self.compList.addRow(
                        self.compileId(player.name, TYPE_COMP),
                        player.name + player.display_titles(),
                        TYPE_COMP,
                    )
                elif "(B)" in str(player.display_titles()):
                    self.blindList.addRow(
                        self.compileId(player.name, TYPE_BLINDFOLD),
                        player.name + player.display_titles(),
                        TYPE_BLINDFOLD,
                    )
                elif "Guest" in str(player.name):
                    self.guestList.addRow(
                        self.compileId(player.name, TYPE_GUEST),
                        player.name + player.display_titles(),
                        TYPE_GUEST,
                    )
                else:
                    self.playersList.addRow(
                        self.compileId(player.name, TYPE_PERSONAL),
                        player.name + player.display_titles(),
                        TYPE_PERSONAL,
                    )
            return False

        self.connection.players.connect("FICSPlayerEntered", addPlayer)

        def removePlayer(players, player):
            if str(player.name) in list(self.connection.notify_users):
                self.friendsList.removeRow(self.compileId(player.name, TYPE_PERSONAL))
            else:
                self.playersList.removeRow(self.compileId(player.name, TYPE_PERSONAL))
            return False

        self.connection.players.connect("FICSPlayerExited", removePlayer)

    def _addChannels(self, channels):
        for grp_id, name in channels:
            grp_id = self.compileId(grp_id, TYPE_CHANNEL)
            self.channelsList.addRow(grp_id, str(grp_id) + ": " + name, TYPE_CHANNEL)

        for grp_id, name in channels:
            if grp_id in self.connection.cm.getJoinedChannels():
                grp_id = self.compileId(grp_id, TYPE_CHANNEL)
                if grp_id.isdigit():
                    self.onAdd(
                        self.channelsList,
                        grp_id,
                        str(grp_id) + ": " + name,
                        TYPE_CHANNEL,
                    )
                else:
                    self.onAdd(self.channelsList, grp_id, name, TYPE_CHANNEL)

    def onChannelsListed(self, cm, channels):
        if not self.channels:
            self.channels = channels
            self._addChannels(channels)

    def compileId(self, grp_id, type):
        if type == TYPE_CHANNEL:
            # FIXME: We can't really add stuff to the grp_id, as panels use it to
            # identify the channel
            assert not grp_id.startswith("person"), "Oops, this is a problem"
        else:
            grp_id = "person" + grp_id.lower()
        return grp_id

    def onAdd(self, grp_list, grp_id, text, grp_type):
        if grp_id in grp_list:
            grp_list.removeRow(grp_id)
        self.joinedList.addRow(grp_id, text, grp_type)
        self.emit("conversationAdded", grp_id, text, grp_type)
        if grp_type == TYPE_CHANNEL:
            self.connection.cm.joinChannel(grp_id)
        self.joinedList.selectRow(grp_id)

    def onRemove(self, joined_list, grp_id, text, grp_type):
        joined_list.removeRow(grp_id)
        if grp_type == TYPE_CHANNEL:
            self.channelsList.addRow(grp_id, text, grp_type)
        elif grp_type == TYPE_PERSONAL:
            self.playersList.addRow(grp_id, text, grp_type)
        elif grp_type == TYPE_COMP:
            self.compList.addRow(grp_id, text, grp_type)
        elif grp_type == TYPE_ADMIN:
            self.adminList.addRow(grp_id, text, grp_type)
        elif grp_type == TYPE_GUEST:
            self.guestList.addRow(grp_id, text, grp_type)
        elif grp_type == TYPE_BLINDFOLD:
            self.blindList.addRow(grp_id, text, grp_type)

        self.emit("conversationRemoved", grp_id)
        if grp_type == TYPE_CHANNEL:
            self.connection.cm.removeChannel(grp_id)

    def onSelect(self, joined_list, grp_id, grp_type):
        self.emit("conversationSelected", grp_id)
        joined_list.get_selection().get_selected()[1]  # Selected iter
        cell = joined_list.leftcol.get_cells()[0]
        self.highlighted[grp_id] = False
        joined_list.leftcol.set_cell_data_func(
            cell, self.change_fg_colour, func_data=self.highlighted
        )

    def onPersonMessage(self, cm, name, title, isadmin, text):
        if not self.compileId(name, TYPE_PERSONAL) in self.joinedList:
            grp_id = self.compileId(name, TYPE_PERSONAL)
            self.onAdd(self.playersList, grp_id, name, TYPE_PERSONAL)



""":Description: This module facilitates configurable object that the end user can
customise such as which chess setor board colours to use or the ability to turn on/off
various sidepanel facilities such as hints, comments engine analysis etc. It also allows
the user to setup and use customised sounds or no sounds at all for a variety of in game
events such as running out of time or piece movement events etc.
"""

import asyncio
import os
from os import listdir
from os.path import isdir, isfile

import sys
from xml.dom import minidom
from urllib.request import url2pathname, pathname2url
from urllib.parse import unquote

from gi.repository import Gtk, GdkPixbuf, Gdk

from pychess.System.prefix import addDataPrefix
from pychess.System import conf, gstreamer, uistuff
from pychess.Players.engineNest import discoverer
from pychess.Utils import book
from pychess.Utils.const import (
    HINT,
    SPY,
    SOUND_MUTE,
    SOUND_BEEP,
    SOUND_URI,
    SOUND_SELECT,
    COUNT_OF_SOUNDS,
)
from pychess.Utils.IconLoader import load_icon, get_pixbuf
from pychess.gfx import Pieces
from pychess.widgets import mainwindow
from pychess.widgets.Background import newTheme
from pychess.perspectives import perspective_manager


firstRun = True

general_tab = None
hint_tab = None
theme_tab = None
sound_tab = None
save_tab = None


def run(widgets):
    global firstRun
    if firstRun:
        initialize(widgets)
        firstRun = False
    else:
        widgets["preferences_dialog"].show()
        widgets["preferences_dialog"].present()


def initialize(widgets):
    """:Description: Initialises the various tabs for each section of configurable artifacts"""
    global general_tab
    general_tab = GeneralTab(widgets)

    # All side panels can be show/hide from View menu now, so no need to do the same from preferences
    # We can re enable this after implementing install/uninstall functionality in the future...
    # PanelTab(widgets)

    uistuff.keepWindowSize("preferencesdialog", widgets["preferences_dialog"])

    notebook = widgets["preferences_notebook"]

    def switch_page(widget, page, page_num):
        global hint_tab, theme_tab, sound_tab, save_tab

        if page_num == 1 and hint_tab is None:
            hint_tab = HintTab(widgets)
        elif page_num == 3 and theme_tab is None:
            theme_tab = ThemeTab(widgets)
        elif page_num == 4 and sound_tab is None:
            sound_tab = SoundTab(widgets)
        elif page_num == 5 and save_tab is None:
            save_tab = SaveTab(widgets)

    notebook.connect("switch_page", switch_page)

    def delete_event(widget, _):
        widgets["preferences_dialog"].hide()
        return True

    widgets["preferences_dialog"].connect("delete-event", delete_event)
    widgets["preferences_dialog"].connect(
        "key-press-event",
        lambda w, e: (
            w.event(Gdk.Event(Gdk.EventType.DELETE))
            if e.keyval == Gdk.KEY_Escape
            else None
        ),
    )


# General initing


class GeneralTab:
    def __init__(self, widgets):
        # Give to uistuff.keeper
        for key in conf.DEFAULTS["General"]:
            # widgets having special getter/setter
            if key in (
                "ana_combobox",
                "inv_ana_combobox",
                "pieceTheme",
                "board_style",
                "board_frame",
            ):
                continue

            try:
                if widgets[key] is not None:
                    uistuff.keep(widgets[key], key)
            except AttributeError:
                print("GeneralTab AttributeError", key, conf.DEFAULTS["General"][key])
            except TypeError:
                print("GeneralTab TypeError", key, conf.DEFAULTS["General"][key])


# Hint initing


def anal_combo_get_value(combobox):
    engine = list(discoverer.getAnalyzers())[combobox.get_active()]
    return engine.get("md5")


def anal_combo_set_value(combobox, value, show_arrow_check, analyzer_type):
    engine = discoverer.getEngineByMd5(value)
    if engine is None:
        combobox.set_active(0)
        # This return saves us from the None-engine being used
        # in later code  -Jonas Thiem
        return
    else:
        try:
            index = list(discoverer.getAnalyzers()).index(engine)
        except ValueError:
            index = 0
        combobox.set_active(index)

    from pychess.widgets.gamewidget import widgets

    perspective = perspective_manager.get_perspective("games")
    for gmwidg in perspective.gamewidgets:
        spectators = gmwidg.gamemodel.spectators
        md5 = engine.get("md5")

        if analyzer_type in spectators and spectators[analyzer_type].md5 != md5:
            gmwidg.gamemodel.remove_analyzer(analyzer_type)
            asyncio.create_task(gmwidg.gamemodel.start_analyzer(analyzer_type))
            if not widgets[show_arrow_check].get_active():
                gmwidg.gamemodel.pause_analyzer(analyzer_type)


class HintTab:
    def __init__(self, widgets):
        self.widgets = widgets

        # Opening book
        path = conf.get("opening_file_entry")
        conf.set("opening_file_entry", path)

        book_chooser_dialog = Gtk.FileChooserDialog(
            _("Select book file"),
            mainwindow(),
            Gtk.FileChooserAction.OPEN,
            (
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OPEN,
                Gtk.ResponseType.OK,
            ),
        )
        book_chooser_button = Gtk.FileChooserButton.new_with_dialog(book_chooser_dialog)

        filter = Gtk.FileFilter()
        filter.set_name(_("Opening books"))
        filter.add_pattern("*.bin")
        book_chooser_dialog.add_filter(filter)
        book_chooser_button.set_filename(path)

        self.widgets["bookChooserDock"].add(book_chooser_button)
        book_chooser_button.show()

        def select_new_book(button):
            new_book = book_chooser_dialog.get_filename()
            if new_book:
                conf.set("opening_file_entry", new_book)
                book.path = new_book
            else:
                # restore the original
                book_chooser_dialog.set_filename(path)

        book_chooser_button.connect("file-set", select_new_book)

        def on_opening_check_toggled(check):
            self.widgets["opening_hbox"].set_sensitive(check.get_active())

        self.widgets["opening_check"].connect_after("toggled", on_opening_check_toggled)

        uistuff.keep(self.widgets["book_depth_max"], "book_depth_max")
        uistuff.keep(self.widgets["book_check_exact_match"], "book_exact_match")

        # Endgame
        egtb_path = conf.get("egtb_path")
        conf.set("egtb_path", egtb_path)

        egtb_chooser_dialog = Gtk.FileChooserDialog(
            _("Select Gaviota TB path"),
            mainwindow(),
            Gtk.FileChooserAction.SELECT_FOLDER,
            (
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OPEN,
                Gtk.ResponseType.OK,
            ),
        )
        egtb_chooser_button = Gtk.FileChooserButton.new_with_dialog(egtb_chooser_dialog)
        egtb_chooser_button.set_current_folder(egtb_path)

        self.widgets["egtbChooserDock"].add(egtb_chooser_button)
        egtb_chooser_button.show()

        def select_egtb(button):
            new_directory = egtb_chooser_dialog.get_filename()
            if new_directory != egtb_path:
                conf.set("egtb_path", new_directory)

        egtb_chooser_button.connect("current-folder-changed", select_egtb)

        def on_endgame_check_toggled(check):
            self.widgets["endgame_hbox"].set_sensitive(check.get_active())

        self.widgets["endgame_check"].connect_after("toggled", on_endgame_check_toggled)

        # Analyzing engines
        from pychess.widgets import newGameDialog

        data = [(item[0], item[1]) for item in newGameDialog.analyzerItems]
        uistuff.createCombo(widgets["ana_combobox"], data, name="ana_combobox")
        uistuff.createCombo(widgets["inv_ana_combobox"], data, name="inv_ana_combobox")

        def update_analyzers_store(discoverer):
            data = [(item[0], item[1]) for item in newGameDialog.analyzerItems]
            uistuff.updateCombo(widgets["ana_combobox"], data)
            uistuff.updateCombo(widgets["inv_ana_combobox"], data)

        discoverer.connect_after("all_engines_discovered", update_analyzers_store)
        update_analyzers_store(discoverer)

        # Save, load and make analyze combos active

        conf.set("ana_combobox", conf.get("ana_combobox"))
        conf.set("inv_ana_combobox", conf.get("inv_ana_combobox"))

        def on_analyzer_check_toggled(check):
            self.widgets["analyzers_vbox"].set_sensitive(check.get_active())
            from pychess.widgets.gamewidget import widgets

            perspective = perspective_manager.get_perspective("games")
            if len(perspective.gamewidgets) != 0:
                if check.get_active():
                    for gmwidg in perspective.gamewidgets:
                        asyncio.create_task(gmwidg.gamemodel.restart_analyzer(HINT))
                        if not widgets["hint_mode"].get_active():
                            gmwidg.gamemodel.pause_analyzer(HINT)
                else:
                    for gmwidg in perspective.gamewidgets:
                        gmwidg.gamemodel.remove_analyzer(HINT)

        self.widgets["analyzers_vbox"].set_sensitive(
            widgets["analyzer_check"].get_active()
        )
        self.widgets["analyzer_check"].connect_after(
            "toggled", on_analyzer_check_toggled
        )

        def on_invanalyzer_check_toggled(check):
            self.widgets["inv_analyzers_vbox"].set_sensitive(check.get_active())
            perspective = perspective_manager.get_perspective("games")
            if len(perspective.gamewidgets) != 0:
                if check.get_active():
                    for gmwidg in perspective.gamewidgets:
                        asyncio.create_task(gmwidg.gamemodel.restart_analyzer(SPY))
                        if not widgets["spy_mode"].get_active():
                            gmwidg.gamemodel.pause_analyzer(SPY)
                else:
                    for gmwidg in perspective.gamewidgets:
                        gmwidg.gamemodel.remove_analyzer(SPY)

        self.widgets["inv_analyzers_vbox"].set_sensitive(
            widgets["inv_analyzer_check"].get_active()
        )
        self.widgets["inv_analyzer_check"].connect_after(
            "toggled", on_invanalyzer_check_toggled
        )

        # Give widgets to keeper

        uistuff.keep(
            self.widgets["ana_combobox"],
            "ana_combobox",
            anal_combo_get_value,
            lambda combobox, value: anal_combo_set_value(
                combobox, value, "hint_mode", HINT
            ),
        )
        uistuff.keep(
            self.widgets["inv_ana_combobox"],
            "inv_ana_combobox",
            anal_combo_get_value,
            lambda combobox, value: anal_combo_set_value(
                combobox, value, "spy_mode", SPY
            ),
        )

        uistuff.keep(self.widgets["max_analysis_spin"], "max_analysis_spin")
        uistuff.keep(self.widgets["infinite_analysis"], "infinite_analysis")
        uistuff.keep(self.widgets["infinite_depth"], "infinite_depth")
        uistuff.keep(self.widgets["max_depth_spin"], "max_depth_spin")


# Sound initing

# Setup default sounds
EXT = "wav" if sys.platform == "win32" else "ogg"

for i in range(COUNT_OF_SOUNDS):
    if not conf.hasKey("soundcombo%d" % i):
        conf.set("soundcombo%d" % i, SOUND_URI)

if not conf.hasKey("sounduri0"):
    conf.set(
        "sounduri0", "file:" + pathname2url(addDataPrefix("sounds/move1.%s" % EXT))
    )
if not conf.hasKey("sounduri1"):
    conf.set(
        "sounduri1", "file:" + pathname2url(addDataPrefix("sounds/check1.%s" % EXT))
    )
if not conf.hasKey("sounduri2"):
    conf.set(
        "sounduri2", "file:" + pathname2url(addDataPrefix("sounds/capture1.%s" % EXT))
    )
if not conf.hasKey("sounduri3"):
    conf.set(
        "sounduri3", "file:" + pathname2url(addDataPrefix("sounds/start1.%s" % EXT))
    )
if not conf.hasKey("sounduri4"):
    conf.set("sounduri4", "file:" + pathname2url(addDataPrefix("sounds/win1.%s" % EXT)))
if not conf.hasKey("sounduri5"):
    conf.set(
        "sounduri5", "file:" + pathname2url(addDataPrefix("sounds/lose1.%s" % EXT))
    )
if not conf.hasKey("sounduri6"):
    conf.set(
        "sounduri6", "file:" + pathname2url(addDataPrefix("sounds/draw1.%s" % EXT))
    )
if not conf.hasKey("sounduri7"):
    conf.set(
        "sounduri7", "file:" + pathname2url(addDataPrefix("sounds/obs_mov.%s" % EXT))
    )
if not conf.hasKey("sounduri8"):
    conf.set(
        "sounduri8", "file:" + pathname2url(addDataPrefix("sounds/obs_end.%s" % EXT))
    )
if not conf.hasKey("sounduri9"):
    conf.set(
        "sounduri9", "file:" + pathname2url(addDataPrefix("sounds/alarm.%s" % EXT))
    )
if not conf.hasKey("sounduri10"):
    conf.set(
        "sounduri10", "file:" + pathname2url(addDataPrefix("sounds/invalid.%s" % EXT))
    )
if not conf.hasKey("sounduri11"):
    conf.set(
        "sounduri11", "file:" + pathname2url(addDataPrefix("sounds/success.%s" % EXT))
    )
if not conf.hasKey("sounduri12"):
    conf.set(
        "sounduri12", "file:" + pathname2url(addDataPrefix("sounds/choice.%s" % EXT))
    )


class SoundTab:
    SOUND_DIRS = (
        addDataPrefix("sounds"),
        "/usr/share/sounds",
        "/usr/local/share/sounds",
        os.path.expanduser("~"),
    )

    actionToKeyNo = {
        "aPlayerMoves": 0,
        "aPlayerChecks": 1,
        "aPlayerCaptures": 2,
        "gameIsSetup": 3,
        "gameIsWon": 4,
        "gameIsLost": 5,
        "gameIsDrawn": 6,
        "observedMoves": 7,
        "oberservedEnds": 8,
        "shortOnTime": 9,
        "invalidMove": 10,
        "puzzleSuccess": 11,
        "variationChoice": 12,
    }

    _player = None

    soundcombo = []
    sounduri = []
    for i in range(COUNT_OF_SOUNDS):
        soundcombo.append(conf.get("soundcombo%s" % i))
        sounduri.append(conf.get("sounduri%s" % i))

    @classmethod
    def getPlayer(cls):
        if not cls._player:
            cls._player = gstreamer.sound_player
        return cls._player

    @classmethod
    def playAction(cls, action):
        if not conf.get("useSounds"):
            return

        if isinstance(action, str):
            key_no = cls.actionToKeyNo[action]
        else:
            key_no = action
        typ = cls.soundcombo[key_no]
        if typ == SOUND_BEEP:
            sys.stdout.write("\a")
            sys.stdout.flush()
        elif typ == SOUND_URI:
            uri = cls.sounduri[key_no]
            if not os.path.isfile(url2pathname(uri[5:])):
                conf.set("soundcombo%d" % key_no, SOUND_MUTE)
                return
            cls.getPlayer().play(uri)

    def __init__(self, widgets):
        # Init open dialog

        opendialog = Gtk.FileChooserDialog(
            _("Open Sound File"),
            mainwindow(),
            Gtk.FileChooserAction.OPEN,
            (
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OPEN,
                Gtk.ResponseType.ACCEPT,
            ),
        )

        for dir in self.SOUND_DIRS:
            if os.path.isdir(dir):
                opendialog.set_current_folder(dir)
                break

        soundfilter = Gtk.FileFilter()
        soundfilter.set_name(_("Sound files"))
        soundfilter.add_mime_type("audio/%s" % EXT)
        soundfilter.add_pattern("*.%s" % EXT)
        opendialog.add_filter(soundfilter)

        # Get combo icons

        icons = (
            (_("No sound"), "audio-volume-muted", "audio-volume-muted"),
            (_("Beep"), "stock_bell", "audio-x-generic"),
            (_("Select sound file..."), "gtk-open", "document-open"),
        )

        items = []
        for level, stock, altstock in icons:
            image = load_icon(16, stock, altstock)
            items += [(image, level)]

        audioIco = load_icon(16, "audio-x-generic")

        # Set-up combos

        def callback(combobox, index):
            if combobox.get_active() == SOUND_SELECT:
                if opendialog.run() == Gtk.ResponseType.ACCEPT:
                    uri = opendialog.get_uri()
                    model = combobox.get_model()
                    conf.set("sounduri%d" % index, uri)
                    self.sounduri[index] = uri
                    label = unquote(os.path.split(uri)[1])
                    if len(model) == 3:
                        model.append([audioIco, label])
                    else:
                        model.set(model.get_iter((3,)), 1, label)
                    combobox.set_active(3)
                else:
                    combobox.set_active(conf.get("soundcombo%d" % index))
                opendialog.hide()

        for i in range(COUNT_OF_SOUNDS):
            combo = widgets["sound%dcombo" % i]
            uistuff.createCombo(combo, items, name="soundcombo%d" % i)
            combo.connect("changed", callback, i)

            label = widgets["soundlabel%d" % i]
            label.props.mnemonic_widget = combo

            uri = conf.get("sounduri%d" % i)
            if os.path.isfile(url2pathname(uri[5:])):
                model = combo.get_model()
                model.append([audioIco, unquote(os.path.split(uri)[1])])

        for i in range(COUNT_OF_SOUNDS):
            if conf.get("soundcombo%d" % i) == SOUND_URI and not os.path.isfile(
                url2pathname(conf.get("sounduri%d" % i)[5:])
            ):
                conf.set("soundcombo%d" % i, SOUND_MUTE)
            uistuff.keep(widgets["sound%dcombo" % i], "soundcombo%d" % i)

        # Init play button

        def playCallback(button, index):
            SoundTab.playAction(index)

        for i in range(COUNT_OF_SOUNDS):
            button = widgets["sound%dbutton" % i]
            button.connect("clicked", playCallback, i)

        # Init 'use sound" checkbutton

        def checkCallBack(*args):
            checkbox = widgets["useSounds"]
            widgets["sounds_frame"].set_property("sensitive", checkbox.get_active())

        conf.notify_add("useSounds", checkCallBack)
        widgets["useSounds"].set_active(conf.get("useSounds"))
        uistuff.keep(widgets["useSounds"], "useSounds")
        checkCallBack()

        if not self.getPlayer().ready:
            widgets["useSounds"].set_sensitive(False)
            widgets["useSounds"].set_active(False)


# Panel initing


class PanelTab:
    def __init__(self, widgets):
        # Put panels in trees
        self.widgets = widgets
        persp = perspective_manager.get_perspective("games")
        sidePanels = persp.sidePanels
        dockLocation = persp.dockLocation

        saved_panels = []
        xmlOK = os.path.isfile(dockLocation)
        if xmlOK:
            doc = minidom.parse(dockLocation)
            for elem in doc.getElementsByTagName("panel"):
                saved_panels.append(elem.getAttribute("id"))

        store = Gtk.ListStore(bool, GdkPixbuf.Pixbuf, str, object)
        for panel in sidePanels:
            checked = True if not xmlOK else panel.__name__ in saved_panels
            panel_icon = get_pixbuf(panel.__icon__, 32)
            text = f"<b>{panel.__title__}</b>\n{panel.__desc__}"
            store.append((checked, panel_icon, text, panel))

        self.tv = widgets["panels_treeview"]
        self.tv.set_model(store)

        self.widgets["panel_about_button"].connect("clicked", self.panel_about)
        self.widgets["panel_enable_button"].connect("toggled", self.panel_toggled)
        self.tv.get_selection().connect("changed", self.selection_changed)

        pixbuf = Gtk.CellRendererPixbuf()
        pixbuf.props.yalign = 0
        pixbuf.props.ypad = 3
        pixbuf.props.xpad = 3
        self.tv.append_column(Gtk.TreeViewColumn("Icon", pixbuf, pixbuf=1, sensitive=0))

        uistuff.appendAutowrapColumn(self.tv, "Name", markup=2, sensitive=0)

        widgets["preferences_notebook"].connect("switch-page", self.__on_switch_page)
        widgets["preferences_dialog"].connect("show", self.__on_show_window)
        widgets["preferences_dialog"].connect("hide", self.__on_hide_window)

    def selection_changed(self, treeselection):
        store, iter = self.tv.get_selection().get_selected()
        self.widgets["panel_enable_button"].set_sensitive(bool(iter))
        self.widgets["panel_about_button"].set_sensitive(bool(iter))

        if iter:
            active = self.tv.get_model().get(iter, 0)[0]
            self.widgets["panel_enable_button"].set_active(active)

    def panel_about(self, button):
        store, iter = self.tv.get_selection().get_selected()
        assert iter  # The button should only be clickable when we have a selection
        path = store.get_path(iter)
        panel = store[path][3]

        d = Gtk.MessageDialog(
            mainwindow(), type=Gtk.MessageType.INFO, buttons=Gtk.ButtonsType.CLOSE
        )
        d.set_markup("<big><b>%s</b></big>" % panel.__title__)
        text = (
            panel.__about__ if hasattr(panel, "__about__") else _("Undescribed panel")
        )
        d.format_secondary_text(text)
        d.run()
        d.hide()

    def panel_toggled(self, button):
        store, iter = self.tv.get_selection().get_selected()
        assert iter  # The button should only be clickable when we have a selection
        path = store.get_path(iter)
        active = button.get_active()
        if store[path][0] == active:
            return

        store[path][0] = active
        self.__set_panel_active(store[path][3], active)

    def __set_panel_active(self, panel, active):
        name = panel.__name__

        from pychess.widgets.pydock import EAST

        persp = perspective_manager.get_perspective("games")
        if active:
            leaf = persp.notebooks["board"].get_parent().get_parent()
            leaf.dock(persp.docks[name][1], EAST, persp.docks[name][0], name)
            panel.menu_item.show()
        else:
            try:
                persp.notebooks[name].get_parent().get_parent().undock(
                    persp.notebooks[name]
                )
                panel.menu_item.hide()
            except AttributeError:
                # A new panel appeared in the panels directory
                leaf = persp.notebooks["board"].get_parent().get_parent()
                leaf.dock(persp.docks[name][1], EAST, persp.docks[name][0], name)

    def showit(self):
        from pychess.widgets.gamewidget import showDesignGW

        showDesignGW()

    def hideit(self):
        from pychess.widgets.gamewidget import hideDesignGW

        hideDesignGW()

    def __on_switch_page(self, notebook, page, page_num):
        if notebook.get_nth_page(page_num) == self.widgets["sidepanels"]:
            self.showit()
        else:
            self.hideit()

    def __on_show_window(self, widget):
        notebook = self.widgets["preferences_notebook"]
        page_num = notebook.get_current_page()
        if notebook.get_nth_page(page_num) == self.widgets["sidepanels"]:
            self.showit()

    def __on_hide_window(self, widget):
        self.hideit()


# Theme initing

board_items = [(None, "colors only")]
boards_path = addDataPrefix("boards")
board_items += [
    (get_pixbuf(os.path.join(boards_path, b), 24), b[:-6])
    for b in listdir(boards_path)
    if b.endswith("_d.png")
]
board_items.append((None, "transparent"))


class ThemeTab:
    """:Description: Allows the setting of various user specific chess
    sets and board colours
    """

    def __init__(self, widgets):
        self.widgets = widgets

        # Font chooser
        font = conf.get("movetextFont")
        font_button = Gtk.FontButton.new_with_font(font)
        demo_text = "♔a1 ♕f8 ♖h8 ♗g7 ♘g2 Ka1 Qf8 Rh8 Bg7 Ng2"
        font_button.set_preview_text(demo_text)
        self.widgets["fontChooserDock"].add(font_button)
        font_button.show()

        def select_font(button):
            conf.set("movetextFont", button.get_font_name())

        font_button.connect("font-set", select_font)

        # Background image
        path = conf.get("welcome_image")
        conf.set("welcome_image", path)

        image_chooser_dialog = Gtk.FileChooserDialog(
            _("Select background image file"),
            mainwindow(),
            Gtk.FileChooserAction.OPEN,
            (
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OPEN,
                Gtk.ResponseType.OK,
            ),
        )
        image_chooser_button = Gtk.FileChooserButton.new_with_dialog(
            image_chooser_dialog
        )

        filter = Gtk.FileFilter()
        filter.set_name(_("Images"))
        filter.add_pattern("*.bmp")
        filter.add_pattern("*.jpg")
        filter.add_pattern("*.png")
        filter.add_pattern("*.svg")
        image_chooser_dialog.add_filter(filter)
        image_chooser_button.set_filename(path)

        self.widgets["imageChooserDock"].add(image_chooser_button)
        image_chooser_button.show()

        def select_new_image(button):
            new_image = image_chooser_dialog.get_filename()
            if new_image:
                conf.set("welcome_image", new_image)
                from pychess.widgets.TaskerManager import tasker

                newTheme(tasker, background=new_image)
                tasker.queue_draw()
            else:
                # restore the original
                image_chooser_dialog.set_filename(path)

        image_chooser_button.connect("file-set", select_new_image)

        # Board style
        uistuff.createCombo(widgets["board_style"], name="board_style")
        data = [(item[0], item[1]) for item in board_items]

        uistuff.createCombo(widgets["board_style"], data)
        uistuff.keep(widgets["board_style"], "board_style")

        # conf.set("board_style", conf.get("board_style"))

        # Board frame
        uistuff.createCombo(widgets["board_frame"], name="board_frame")
        data = [(item[0], item[1]) for item in [(None, "no frame")] + board_items[1:]]

        uistuff.createCombo(widgets["board_frame"], data)
        uistuff.keep(widgets["board_frame"], "board_frame")

        # conf.set("board_frame", conf.get("board_frame"))

        # Board Colours

        def onColourSetLight(_):
            """:Description: Sets the light squares of the chess board
            to the value selected in the colour picker
            """
            conf.set("lightcolour", widgets["light_cbtn"].get_color().to_string())

        widgets["light_cbtn"].connect_after("color-set", onColourSetLight)

        def onColourSetDark(_):
            """:Description: Sets the dark squares of the chess board
            to the value selected in the colour picker
            """
            conf.set("darkcolour", widgets["dark_cbtn"].get_color().to_string())

        widgets["dark_cbtn"].connect_after("color-set", onColourSetDark)

        def onResetColourClicked(_):
            """:Description: Resets the chess board squares to factory default"""
            conf.set("lightcolour", conf.DEFAULTS["General"]["lightcolour"])
            conf.set("darkcolour", conf.DEFAULTS["General"]["darkcolour"])

        widgets["reset_btn"].connect("clicked", onResetColourClicked)

        # Get the current board colours if set, if not set, set them to default
        conf.set("lightcolour", conf.get("lightcolour"))
        conf.set("darkcolour", conf.get("darkcolour"))

        # Next 2 lines take a #hex str converts them to a color then to a RGBA representation
        self.lightcolour = Gdk.RGBA()
        self.lightcolour.parse(conf.get("lightcolour"))
        self.darkcolour = Gdk.RGBA()
        self.darkcolour.parse(conf.get("darkcolour"))

        # Set the color swatches in preference to stored values
        widgets["light_cbtn"].set_rgba(self.lightcolour)
        widgets["dark_cbtn"].set_rgba(self.darkcolour)

        # Chess Sets

        self.themes = self.discoverThemes()
        store = Gtk.ListStore(GdkPixbuf.Pixbuf, str)

        for theme in self.themes:
            pngfile = "{}/{}.png".format(addDataPrefix("pieces"), theme)

            if isfile(pngfile):
                pixbuf = get_pixbuf(pngfile)
                store.append((pixbuf, theme))
            else:
                print(
                    "WARNING: No piece theme preview icons found. Please run \
                    create_theme_preview.sh !"
                )
                break

        self.icon_view = widgets["pieceTheme"]
        self.icon_view.set_model(store)
        self.icon_view.set_pixbuf_column(0)
        self.icon_view.set_text_column(1)

        def keepSize(crt, _):
            """ :Description: Hack to fix spacing problem in iconview
                http://stackoverflow.com/questions/14090094/what-causes-the-different-\
                    display-behaviour-for-a-gtkiconview-between-different
            """
            crt.handler_block(crt_notify)
            crt.set_property("width", 40)
            crt.handler_unblock(crt_notify)

        crt = self.icon_view.get_cells()[0]
        crt_notify = crt.connect("notify", keepSize)

        def _getActive(iconview):
            model = iconview.get_model()
            selected = iconview.get_selected_items()

            if len(selected) == 0:
                return conf.get("pieceTheme")

            indices = selected[0].get_indices()
            if indices:
                idx = indices[0]
                theme = model[idx][1]
                Pieces.set_piece_theme(theme)
                return theme

        def _setActive(iconview, value):
            try:
                index = self.themes.index(value)
            except ValueError:
                index = 0
            iconview.select_path(
                Gtk.TreePath(
                    index,
                )
            )

        uistuff.keep(widgets["pieceTheme"], "pieceTheme", _getActive, _setActive)

    def discoverThemes(self):
        """:Description: Finds all the different chess sets that are present
        in the pieces directory

        :return: (a List) of themes
        """
        themes = []

        pieces = addDataPrefix("pieces")
        themes += [
            d.capitalize() for d in listdir(pieces) if isdir(os.path.join(pieces, d))
        ]

        themes.sort()

        return themes


# Save initing


class SaveTab:
    """:Description: Allows the user to configure the structure of saved
    game files name along with various game attributes such as elapse time
    between moves and analysis engin evalutations
    """

    def __init__(self, widgets):
        # Init 'auto save" checkbutton
        def checkCallBack(_):
            """:Description: Sets the various option based on user interaction with the
            checkboxes in the gui
            """

            checkbox = widgets["autoSave"]
            widgets["autosave_grid"].set_property("sensitive", checkbox.get_active())

        conf.notify_add("autoSave", checkCallBack)
        uistuff.keep(widgets["autoSave"], "autoSave")
        checkCallBack(_)

        self.auto_save_path = conf.get("autoSavePath")
        conf.set("autoSavePath", self.auto_save_path)

        auto_save_chooser_dialog = Gtk.FileChooserDialog(
            _("Select auto save path"),
            mainwindow(),
            Gtk.FileChooserAction.SELECT_FOLDER,
            (
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OPEN,
                Gtk.ResponseType.OK,
            ),
        )
        auto_save_chooser_button = Gtk.FileChooserButton.new_with_dialog(
            auto_save_chooser_dialog
        )
        auto_save_chooser_button.set_current_folder(self.auto_save_path)

        widgets["savePathChooserDock"].add(auto_save_chooser_button)
        auto_save_chooser_button.show()

        def selectAutoSave(_):
            """:Description: Sets the auto save path for stored games if it
            has changed since last time

            :signal: Activated on receiving the 'current-folder-changed' signal
            """
            new_directory = auto_save_chooser_dialog.get_filename()
            if new_directory != self.auto_save_path:
                conf.set("autoSavePath", new_directory)

        auto_save_chooser_button.connect("current-folder-changed", selectAutoSave)

        conf.set("autoSaveFormat", conf.get("autoSaveFormat"))
        uistuff.keep(widgets["autoSaveFormat"], "autoSaveFormat")

        # uistuff.keep(widgets["saveEmt"], "saveEmt")
        # uistuff.keep(widgets["saveEval"], "saveEval")
        # uistuff.keep(widgets["saveRatingChange"], "saveRatingChange")
        # uistuff.keep(widgets["indentPgn"], "indentPgn")
        # uistuff.keep(widgets["saveOwnGames"], "saveOwnGames")



import cairo

from gi.repository import Gtk, GObject, Gdk

from .OverlayWindow import OverlayWindow

from .__init__ import NORTH, EAST, SOUTH, WEST


class ArrowButton(OverlayWindow):
    """Leafs will connect to the drag-drop signal"""

    __gsignals__ = {
        "dropped": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "hovered": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "left": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self, parent, svgPath, position):
        OverlayWindow.__init__(self, parent)
        self.myposition = position
        self.svgPath = svgPath

        # targets = [("GTK_NOTEBOOK_TAB", Gtk.TargetFlags.SAME_APP, 0xbadbeef)]
        targets = [
            Gtk.TargetEntry.new("GTK_NOTEBOOK_TAB", Gtk.TargetFlags.SAME_APP, 0xBADBEEF)
        ]
        self.drag_dest_set(
            Gtk.DestDefaults.DROP | Gtk.DestDefaults.MOTION,
            targets,
            Gdk.DragAction.MOVE,
        )
        self.drag_dest_set_track_motion(True)
        self.myparent.button_cids[self] += [
            self.connect("drag-motion", self.__onDragMotion),
            self.connect("drag-leave", self.__onDragLeave),
            self.connect("drag-drop", self.__onDragDrop),
            self.connect_after("draw", self.__onExposeEvent),
        ]
        self.hovered = False

        self.myparentAlloc = None
        self.myparentPos = None
        self.hasHole = False

    def _calcSize(self):
        parentAlloc = self.myparent.get_allocation()
        width, height = self.getSizeOfSvg(self.svgPath)

        if self.myparentAlloc is None:
            self.resize(width, height)

        if self.get_window() and not self.hasHole:
            self.hasHole = True
            self.digAHole(self.svgPath, width, height)

        if self.myposition == NORTH:
            x_loc, y_loc = parentAlloc.width / 2.0 - width / 2.0, 0
        elif self.myposition == EAST:
            x_loc, y_loc = (
                parentAlloc.width - width,
                parentAlloc.height / 2.0 - height / 2.0,
            )
        elif self.myposition == SOUTH:
            x_loc, y_loc = (
                parentAlloc.width / 2.0 - width / 2.0,
                parentAlloc.height - height,
            )
        elif self.myposition == WEST:
            x_loc, y_loc = 0, parentAlloc.height / 2.0 - height / 2.0

        x_loc, y_loc = self.translateCoords(int(x_loc), int(y_loc))
        if (x_loc, y_loc) != self.get_position():
            self.move(x_loc, y_loc)

        self.myparentAlloc = parentAlloc
        window = self.myparent.get_window()
        if window is None:
            print("   !!! get_window() returned None for", self.myparent)
        else:
            self.myparentPos = window.get_position()

    def __onExposeEvent(self, self_, ctx):
        self._calcSize()
        context = self.get_window().cairo_create()
        width, height = self.getSizeOfSvg(self.svgPath)
        surface = self.getSurfaceFromSvg(self.svgPath, width, height)

        if self.is_composited():
            context.set_operator(cairo.OPERATOR_CLEAR)
            context.set_source_rgba(0.0, 0.0, 0.0, 0.0)
            context.paint()
            context.set_operator(cairo.OPERATOR_OVER)

        # FIXME
        # mask = Gdk.Pixmap(None, width, height, 1)
        # mcontext = mask.cairo_create()
        # mcontext.set_source_surface(surface, 0, 0)
        # mcontext.paint()
        # self.window.shape_combine_mask(mask, 0, 0)

        context.set_source_surface(surface, 0, 0)
        context.paint()

    def __containsPoint(self, x, y):
        alloc = self.get_allocation()
        return 0 <= x < alloc.width and 0 <= y < alloc.height

    def __onDragMotion(self, arrow, context, x, y, timestamp):
        if not self.hovered and self.__containsPoint(x, y):
            self.hovered = True
            self.emit("hovered", Gtk.drag_get_source_widget(context))
        elif self.hovered and not self.__containsPoint(x, y):
            self.hovered = False
            self.emit("left")

    def __onDragLeave(self, arrow, context, timestamp):
        if self.hovered:
            self.hovered = False
            self.emit("left")

    def __onDragDrop(self, arrow, context, x, y, timestamp):
        if self.__containsPoint(x, y):
            self.emit("dropped", Gtk.drag_get_source_widget(context))
            context.finish(True, True, timestamp)
            return True



import datetime

from gi.repository import GObject

from pychess.ic import GAME_TYPES_BY_FICS_NAME
from pychess.ic.icc import DG_GAMELIST_BEGIN, DG_GAMELIST_ITEM, DG_RATING_TYPE_KEY
from pychess.ic.managers.AdjournManager import AdjournManager
from pychess.ic.FICSObjects import FICSAdjournedGame, FICSHistoryGame, FICSJournalGame
from pychess.Utils.const import (
    WON_ADJUDICATION,
    DRAW_AGREE,
    WON_DISCONNECTION,
    WON_CALLFLAG,
    WON_MATE,
    DRAW_INSUFFICIENT,
    DRAW_REPETITION,
    WON_RESIGN,
    DRAW_STALEMATE,
    DRAW_BLACKINSUFFICIENTANDWHITETIME,
    UNKNOWN_REASON,
    DRAW_50MOVES,
    WHITEWON,
    DRAW,
    BLACKWON,
    ADJOURNED,
    DRAW_CALLFLAG,
    DRAW_ADJUDICATION,
    ABORTED,
    WHITE,
    BLACK,
)

won_reasons_dict = {
    "0": WON_RESIGN,
    "1": WON_MATE,
    "2": WON_CALLFLAG,
    "3": WON_ADJUDICATION,
    "4": WON_DISCONNECTION,
    "5": WON_DISCONNECTION,
    "6": WON_DISCONNECTION,
    "7": WON_RESIGN,
    "8": WON_MATE,
    "9": WON_CALLFLAG,
    "10": WON_DISCONNECTION,
    "11": WON_DISCONNECTION,
    "12": WON_DISCONNECTION,
    "13": WON_ADJUDICATION,
}

draw_reasons_dict = {
    "0": DRAW_AGREE,
    "1": DRAW_STALEMATE,
    "2": DRAW_REPETITION,
    "3": DRAW_50MOVES,
    "4": DRAW_BLACKINSUFFICIENTANDWHITETIME,  # DRAW_WHITEINSUFFICIENTANDBLACKTIME
    "5": DRAW_INSUFFICIENT,
    "6": DRAW_CALLFLAG,
    "7": DRAW_ADJUDICATION,
    "8": DRAW_AGREE,
    "9": DRAW_CALLFLAG,
    "10": DRAW_ADJUDICATION,
}


class ICCAdjournManager(AdjournManager):
    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection

        self.RATING_TYPES = {}

        self.connection.expect_dg_line(DG_RATING_TYPE_KEY, self.on_icc_rating_type_key)
        self.connection.expect_dg_line(DG_GAMELIST_BEGIN, self.on_icc_gamelist_begin)
        self.connection.expect_dg_line(DG_GAMELIST_ITEM, self.on_icc_gamelist_item)

        self.connection.client.run_command("set-2 %s 1" % DG_RATING_TYPE_KEY)
        self.connection.client.run_command("set-2 %s 1" % DG_GAMELIST_BEGIN)
        self.connection.client.run_command("set-2 %s 1" % DG_GAMELIST_ITEM)

        self.queryAdjournments()
        self.queryHistory()
        self.queryJournal()

        self.connection.query_game = None

    def on_icc_rating_type_key(self, data):
        key, value = data.split()
        self.RATING_TYPES[key] = value.lower()

    def on_icc_gamelist_begin(self, data):
        # command {parameters} nhits first last {summary}
        # command is one of search, history, liblist, or stored
        # history {gbtami} 20 1 20 {Recent games of gbtami}
        command, rest = data.split(" ", 1)
        params, rest = rest.split("}", 1)
        name = params[1:]

        # force clean up old gamelists
        if self.connection.history_owner != name:
            self.emit("onAdjournmentsList", [])
            self.emit("onHistoryList", [])
            self.emit("onJournalList", [])

        if command == "stored":
            self.connection.stored_owner = name
        elif command == "history":
            self.connection.history_owner = name
        elif command == "liblist":
            self.connection.journal_owner = name

        self.gamelist_command = command

    def on_icc_gamelist_item(self, data):
        # index id event date time white-name white-rating black-name black-rating rated rating-type
        # wild init-time-W inc-W init-time-B inc-B eco status color mode {note} here
        # status: 0=win 1=draw 2=adjourned 3=abort
        # rating_type: 0=wild, 1=blitz, 2=standard, bullet, 4=bughouse see DG_RATING_TYPE_KEY
        # 99 1731753309 ? 2016.12.08 15:07:53 gbtami 1538 konechno 1644 1 11 0 3 0 3 0 A00 0 1 1 {} 0
        # 98 1731751094 ? 2016.12.08 14:34:37 gbtami 1550 espilva 1484 1 11 0 3 0 3 0 A00 1 0 5 {} 0
        (
            idx,
            gid,
            event,
            date,
            time,
            wname,
            wrating,
            bname,
            brating,
            rated,
            rating_type,
            wild,
            wtime,
            winc,
            btime,
            binc,
            eco,
            status,
            color,
            mode,
            rest,
        ) = data.split(" ", 20)

        if status == "0":
            result = WHITEWON if color == "0" else BLACKWON
        elif status == "1":
            result = DRAW
        elif status == "2":
            result = ADJOURNED
        else:
            result = ABORTED

        white = wname
        black = bname
        wrating = wrating
        brating = brating

        if status == "0":
            reason = won_reasons_dict[mode]
        elif status == "1":
            reason = draw_reasons_dict[mode]
        else:
            reason = UNKNOWN_REASON

        year, month, day = date.split(".")
        hour, minute, sec = time.split(":")
        gametime = datetime.datetime(
            int(year), int(month), int(day), int(hour), int(minute), int(sec)
        )
        rated = rated == "1"
        fics_name = self.RATING_TYPES[rating_type]
        gametype = GAME_TYPES_BY_FICS_NAME[fics_name]

        minutes = int(wtime)
        gain = int(winc)

        wplayer = self.connection.players.get(white)
        bplayer = self.connection.players.get(black)

        if self.gamelist_command == "stored":
            game = FICSAdjournedGame(
                wplayer,
                bplayer,
                game_type=gametype,
                rated=rated,
                our_color=WHITE if wname == self.connection.username else BLACK,
                minutes=minutes,
                inc=gain,
            )

            if game.opponent.adjournment is False:
                game.opponent.adjournment = True

        elif self.gamelist_command == "history":
            game = FICSHistoryGame(
                wplayer,
                bplayer,
                game_type=gametype,
                rated=rated,
                minutes=minutes,
                inc=gain,
                wrating=wrating,
                brating=brating,
                time=gametime,
                reason=reason,
                history_no=idx,
                result=result,
            )

        elif self.gamelist_command == "liblist":
            game = FICSJournalGame(
                wplayer,
                bplayer,
                game_type=gametype,
                rated=rated,
                minutes=minutes,
                inc=gain,
                wrating=wrating,
                brating=brating,
                time=gametime,
                reason=reason,
                journal_no=idx,
                result=result,
            )

        if game not in self.connection.games:
            game = self.connection.games.get(game, emit=False)
            if self.gamelist_command == "stored":
                self.emit("adjournedGameAdded", game)
            elif self.gamelist_command == "history":
                self.emit("historyGameAdded", game)
            elif self.gamelist_command == "liblist":
                self.emit("journalGameAdded", game)

    def queryJournal(self, owner=None):
        if owner is None:
            self.connection.client.run_command("liblist")
        else:
            self.connection.client.run_command("liblist %s" % owner)

    def queryMoves(self, game):
        self.connection.query_game = game
        if isinstance(game, FICSHistoryGame):
            self.connection.client.run_command(
                f"spgn {self.connection.history_owner} {game.history_no}"
            )
        elif isinstance(game, FICSJournalGame):
            self.connection.client.run_command(
                f"spgn {self.connection.journal_owner} %{game.journal_no}"
            )
        else:
            self.connection.client.run_command(
                f"spgn {self.connection.stored_owner} {game.opponent.name}"
            )



# Bughouse Chess

from pychess.Utils.const import VARIANTS_OTHER_NONSTANDARD, BUGHOUSECHESS
from pychess.Utils.Board import Board


class BughouseBoard(Board):
    variant = BUGHOUSECHESS
    __desc__ = _("FICS bughouse: http://www.freechess.org/Help/HelpFiles/bughouse.html")
    name = _("Bughouse")
    cecp_name = "bughouse"
    need_initial_board = False
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD



from gi.repository import Gtk, GObject


class Accordion(Gtk.TreeView):
    def __init__(self, model):
        GObject.GObject.__init__(self, model)

        self.set_headers_visible(False)
        self.set_property("show-expanders", False)
        self.set_property("level-indentation", 10)

        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn("Column", renderer)

        def top_level(column, cell, store, iter):
            cell.set_property("text", store[iter][0])
            if store.iter_depth(iter) == 0:
                cell.set_property("foreground", "black")
                cell.set_property("background", "gray")
            else:
                cell.set_property("foreground", "black")
                cell.set_property("background", "white")

        column.set_cell_data_func(renderer, top_level)

        self.append_column(column)

        selection = self.get_selection()
        selection.set_mode(Gtk.SelectionMode.SINGLE)
        selection.connect("changed", self.on_selection_changed)
        self.current = None

    def on_selection_changed(self, selection, data=None):
        model, iter = selection.get_selected()

        if model.iter_depth(iter) == 0 and iter != self.current:
            self.collapse_all()
            self.expand_row(model.get_path(iter), True)
            self.current = iter

        selected_item = model.get_value(iter, 0)
        print(selected_item)


if __name__ == "__main__":
    window = Gtk.Window(Gtk.WindowType.TOPLEVEL)
    window.set_title("Accordion example")
    window.set_size_request(200, 200)
    window.connect("delete_event", Gtk.main_quit)

    treestore = Gtk.TreeStore(str)
    for parent in range(4):
        piter = treestore.append(None, ["parent %i" % parent])
        for child in range(3):
            treestore.append(piter, ["child %i of parent %i" % (child, parent)])
    accordion = Accordion(treestore)
    window.add(accordion)
    window.show_all()
    Gtk.main()



import asyncio
import re
import sys
import socket
import traceback
from collections import defaultdict
from concurrent.futures import CancelledError

from gi.repository import GObject

import pychess
from pychess.System.Log import log

from pychess import ic
from pychess.Utils.const import (
    NAME,
    FISCHERRANDOMCHESS,
    LOSERSCHESS,
    ATOMICCHESS,
    CRAZYHOUSECHESS,
)
from .managers.SeekManager import SeekManager
from .managers.FingerManager import FingerManager
from .managers.NewsManager import NewsManager
from .managers.BoardManager import BoardManager
from .managers.OfferManager import OfferManager
from .managers.ChatManager import ChatManager
from .managers.ConsoleManager import ConsoleManager
from .managers.HelperManager import HelperManager
from .managers.ListAndVarManager import ListAndVarManager
from .managers.AutoLogOutManager import AutoLogOutManager
from .managers.ErrorManager import ErrorManager
from .managers.AdjournManager import AdjournManager
from .managers.ICCSeekManager import ICCSeekManager
from .managers.ICCBoardManager import ICCBoardManager
from .managers.ICCChatManager import ICCChatManager
from .managers.ICCHelperManager import ICCHelperManager
from .managers.ICCAdjournManager import ICCAdjournManager
from .managers.ICCErrorManager import ICCErrorManager
from .managers.ICCFingerManager import ICCFingerManager
from .managers.ICCListAndVarManager import ICCListAndVarManager
from .managers.ICCNewsManager import ICCNewsManager
from .managers.ICCOfferManager import ICCOfferManager
from .managers.ICCAutoLogOutManager import ICCAutoLogOutManager

from .FICSObjects import FICSPlayers, FICSGames, FICSSeeks, FICSChallenges
from .TimeSeal import CanceledException, ICSTelnet
from .VerboseTelnet import (
    LinePrediction,
    FromPlusPrediction,
    FromABPlusPrediction,
    FromToPrediction,
    PredictionsTelnet,
    NLinesPrediction,
)


class LogOnException(Exception):
    pass


class Connection(GObject.GObject):
    __gsignals__ = {
        "connecting": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "connectingMsg": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        "connected": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "disconnected": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "error": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
    }

    def __init__(self, host, ports, timeseal, username, password):
        GObject.GObject.__init__(self)
        self.daemon = True
        self.host = host
        self.ports = ports
        self.username = username
        self.password = password
        self.timeseal = timeseal

        self.connected = False
        self.connecting = False
        self.keep_alive_task = None

        self.predictions = set()
        self.predictionsDict = {}
        self.reply_cmd_dict = defaultdict(list)

        # Are we connected to FatICS ?
        self.FatICS = False

        self.USCN = False

        self.ICC = False
        self.replay_dg_dict = {}
        self.replay_cn_dict = {}

    @property
    def ics_name(self):
        if self.FatICS:
            return "FatICS"
        elif self.USCN:
            return "USCN"
        elif self.ICC:
            return "ICC"
        else:
            return "FICS"

    def expect(self, prediction):
        self.predictions.add(prediction)
        self.predictionsDict[prediction.callback] = prediction
        if hasattr(prediction.callback, "BLKCMD"):
            predictions = self.reply_cmd_dict[prediction.callback.BLKCMD]
            predictions.append(prediction)
            # Do reverse sorted so we can later test the longest predictions first.
            # This is so that matching prefers the longest match for matches
            # that start out with the same regexp line(s)
            self.reply_cmd_dict[prediction.callback.BLKCMD] = sorted(
                predictions, key=len, reverse=True
            )

    def unexpect(self, callback):
        self.predictions.remove(self.predictionsDict.pop(callback))
        if hasattr(callback, "BLKCMD"):
            for prediction in self.reply_cmd_dict[callback.BLKCMD]:
                if prediction.callback == callback:
                    self.reply_cmd_dict[callback.BLKCMD].remove(prediction)
            if len(self.reply_cmd_dict[callback.BLKCMD]) == 0:
                del self.reply_cmd_dict[callback.BLKCMD]

    def expect_dg_line(self, number, callback):
        self.replay_dg_dict[number] = callback

    def expect_cn_line(self, number, callback):
        self.replay_cn_dict[number] = callback

    def expect_line(self, callback, regexp):
        self.expect(LinePrediction(callback, regexp))

    def expect_n_lines(self, callback, *regexps):
        self.expect(NLinesPrediction(callback, *regexps))

    def expect_fromplus(self, callback, regexp0, regexp1):
        self.expect(FromPlusPrediction(callback, regexp0, regexp1))

    def expect_fromABplus(self, callback, regexp0, regexp1, regexp2):
        self.expect(FromABPlusPrediction(callback, regexp0, regexp1, regexp2))

    def expect_fromto(self, callback, regexp0, regexp1):
        self.expect(FromToPrediction(callback, regexp0, regexp1))

    def cancel(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def getUsername(self):
        return self.username

    def isRegistred(self):
        return self.password is not None and self.password != ""

    def isConnected(self):
        return self.connected

    def isConnecting(self):
        return self.connecting


EOF = _('The connection was broken - got "end of file" message')
NOTREG = _("'%s' is not a registered name")
BADPAS = _(
    "The entered password was invalid.\n"
    + "If you forgot your password, go to "
    + '<a href="http://www.freechess.org/password">'
    + "http://www.freechess.org/password</a> to request a new one over email."
)
ALREADYIN = _("Sorry '%s' is already logged in")
REGISTERED = _("'%s' is a registered name.  If it is yours, type the password.")
PREVENTED = _(
    "Due to abuse problems, guest connections have been prevented.\n"
    + "You can still register on http://www.freechess.org"
)
TOOSHORT = _("A name should be at least three characters long!")
LETTERSONLY = _("Sorry, names can only consist of lower and upper case letters.")


class FICSConnection(Connection):
    def __init__(self, host, ports, timeseal, username="guest", password=""):
        Connection.__init__(self, host, ports, timeseal, username, password)

    def _post_connect_hook(self, lines):
        pass

    def _start_managers(self):
        pass

    async def _connect(self):
        self.connecting = True
        self.emit("connecting")
        try:
            self.emit("connectingMsg", _("Connecting to server"))
            for i, port in enumerate(self.ports):
                log.debug("Trying port %d" % port, extra={"task": (self.host, "raw")})
                try:
                    connected_event = asyncio.Event()
                    self.client = ICSTelnet(self.timeseal)
                    asyncio.create_task(
                        self.client.start(self.host, port, connected_event)
                    )
                    await connected_event.wait()
                except OSError as err:
                    log.debug(
                        "Failed to open port %d %s" % (port, err),
                        extra={"task": (self.host, "raw")},
                    )
                    if i + 1 == len(self.ports):
                        raise
                    else:
                        continue
                else:
                    break

            await self.client.read_until("login: ")
            self.emit("connectingMsg", _("Logging on to server"))

            # login with registered handle
            if self.password:
                self.client.write(self.username)
                got = await self.client.read_until(
                    "password:", "enter the server as", "Try again."
                )
                if got == 0:
                    self.client.sensitive = True
                    self.client.write(self.password)
                # No such name
                elif got == 1:
                    raise LogOnException(NOTREG % self.username)
                # Bad name
                elif got == 2:
                    raise LogOnException(NOTREG % self.username)
            else:
                if self.username:
                    self.client.write(self.username)
                else:
                    self.client.write("guest")
                got = await self.client.read_until(
                    "Press return",
                    "You are connected as a guest",
                    "If it is yours, type the password.",
                    "A name should be at least three characters long!",
                    "Sorry, names can only consist of lower and upper case letters.",
                    "guest connections have been prevented",
                    "nobody from your site may login without an account.",
                )
                # got = 3
                if got == 2:
                    raise LogOnException(REGISTERED % self.username)
                elif got == 3:
                    raise LogOnException(TOOSHORT)
                elif got == 4:
                    raise LogOnException(LETTERSONLY)
                elif got == 5 or got == 6:
                    raise LogOnException(PREVENTED)
                self.client.write("")

            while True:
                line = await self.client.readline()
                if "Invalid password" in line:
                    raise LogOnException(BADPAS)
                elif "is already logged in" in line:
                    raise LogOnException(ALREADYIN % self.username)

                match = re.search(
                    r"\*\*\*\* Starting FICS session as (%s)%s \*\*\*\*"
                    % (ic.NAMES_RE, ic.TITLES_RE),
                    line,
                )
                if match:
                    self.username = match.groups()[0]
                    break

                # USCN specific line
                match = re.search("Created temporary login '(%s)'" % ic.NAMES_RE, line)
                if match:
                    self.username = match.groups()[0]
                    break

                match = re.search("For a list of events, click here:", line)
                if match:
                    break

                # ICC specific line
                match = re.search("help anonymous", line)
                if match:
                    break

                match = re.search("This is the admin message of the day", line)
                if match:
                    break

            self.emit("connectingMsg", _("Setting up environment"))
            lines = await self.client.readuntil(b"ics%")
            self._post_connect_hook(lines)
            self.FatICS = self.client.FatICS
            self.USCN = self.client.USCN
            self.ICC = self.client.ICC
            self.client.name = self.username
            self.client = PredictionsTelnet(
                self.client,
                self.predictions,
                self.reply_cmd_dict,
                self.replay_dg_dict,
                self.replay_cn_dict,
            )
            self.client.lines.line_prefix = "aics%" if self.ICC else "fics%"

            if not self.USCN and not self.ICC:
                self.client.run_command("iset block 1")
                self.client.lines.block_mode = True

            if self.ICC:
                self.client.run_command("set level1 5")
                self.client.run_command("set prompt 0")
                self.client.lines.datagram_mode = True

                ic.GAME_TYPES_BY_SHORT_FICS_NAME["B"] = ic.GAME_TYPES["bullet"]
                ic.VARIANT_GAME_TYPES[ATOMICCHESS] = ic.GAME_TYPES["w27"]
                ic.VARIANT_GAME_TYPES[CRAZYHOUSECHESS] = ic.GAME_TYPES["w23"]
                ic.VARIANT_GAME_TYPES[LOSERSCHESS] = ic.GAME_TYPES["w17"]
                ic.VARIANT_GAME_TYPES[FISCHERRANDOMCHESS] = ic.GAME_TYPES["w22"]
            else:
                ic.GAME_TYPES_BY_SHORT_FICS_NAME["B"] = ic.GAME_TYPES["bughouse"]
                ic.VARIANT_GAME_TYPES[ATOMICCHESS] = ic.GAME_TYPES["atomic"]
                ic.VARIANT_GAME_TYPES[CRAZYHOUSECHESS] = ic.GAME_TYPES["crazyhouse"]
                ic.VARIANT_GAME_TYPES[LOSERSCHESS] = ic.GAME_TYPES["losers"]
                ic.VARIANT_GAME_TYPES[FISCHERRANDOMCHESS] = ic.GAME_TYPES["wild/fr"]

            self.client.run_command("iset defprompt 1")
            self.client.run_command("iset ms 1")

            self._start_managers(lines)
            self.connecting = False
            self.connected = True
            self.emit("connected")

            async def keep_alive():
                while self.isConnected():
                    self.client.run_command("date")
                    await asyncio.sleep(30 * 60)

            self.keep_alive_task = asyncio.create_task(keep_alive())

        except CanceledException as err:
            log.info(
                "FICSConnection._connect: %s" % repr(err),
                extra={"task": (self.host, "raw")},
            )
        finally:
            self.connecting = False

    async def start(self):
        try:
            if not self.isConnected():
                await self._connect()
            while self.isConnected():
                await self.client.parse()
        except CancelledError:
            pass
        except Exception as err:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback)
            log.info(
                "FICSConnection.run: %s" % repr(err), extra={"task": (self.host, "raw")}
            )
            self.close()
            if isinstance(
                err,
                (
                    IOError,
                    LogOnException,
                    EOFError,
                    socket.error,
                    socket.gaierror,
                    socket.herror,
                ),
            ):
                self.emit("error", err)
            else:
                raise
        finally:
            if isinstance(self, FICSMainConnection):
                self.emit("disconnected")

    def cancel(self):
        self.close()
        if hasattr(self, "client"):
            self.client.cancel()

    def close(self):
        self.connected = False
        if hasattr(self, "client"):
            self.client.close()
        if self.keep_alive_task is not None:
            self.keep_alive_task.cancel()


class FICSMainConnection(FICSConnection):
    def __init__(self, host, ports, timeseal, username="guest", password=""):
        FICSConnection.__init__(self, host, ports, timeseal, username, password)
        self.lvm = None
        self.notify_users = []
        self.ini_messages = []
        self.players = FICSPlayers(self)
        self.games = FICSGames(self)
        self.seeks = FICSSeeks(self)
        self.challenges = FICSChallenges(self)
        self.archived_examine = None
        self.examined_game = None
        self.stored_owner = self.username
        self.history_owner = self.username
        self.journal_owner = self.username
        self.set_user_vars = False

    def close(self):
        if isinstance(self.client, PredictionsTelnet) and self.set_user_vars:
            self.client.run_command("set open 0")
            self.client.run_command("set gin 0")
            self.client.run_command("set availinfo 0")
        try:
            self.lvm.stop()
        except AttributeError:
            pass
        except Exception as err:
            if not isinstance(
                err,
                (
                    IOError,
                    LogOnException,
                    EOFError,
                    socket.error,
                    socket.gaierror,
                    socket.herror,
                ),
            ):
                raise
        finally:
            FICSConnection.close(self)

    def _post_connect_hook(self, lines):
        self.ini_messages = lines.splitlines()
        notify_users = re.search(
            r"Present company includes: ((?:%s ?)+)\." % ic.NAMES_RE, lines
        )
        if notify_users:
            self.notify_users.extend(notify_users.groups()[0].split())

    def _start_managers(self, lines):
        self.client.run_command(f"set interface {NAME} {pychess.VERSION}")

        # FIXME: Some managers use each other to avoid regexp collapse. To
        # avoid having to init the in a specific order, connect calls should
        # be moved to a "start" function, so all managers would be in
        # the connection object when they are called
        if self.ICC:
            self.lvm = ICCListAndVarManager(self)
            self.em = ICCErrorManager(self)
            self.glm = ICCSeekManager(self)
            self.bm = ICCBoardManager(self)
            self.cm = ICCChatManager(self)
            self.adm = ICCAdjournManager(self)
            self.fm = ICCFingerManager(self)
            self.nm = ICCNewsManager(self)
            self.om = ICCOfferManager(self)
            self.alm = ICCAutoLogOutManager(self)
        else:
            self.lvm = ListAndVarManager(self)
            self.em = ErrorManager(self)
            self.glm = SeekManager(self)
            self.bm = BoardManager(self)
            self.cm = ChatManager(self)
            self.adm = AdjournManager(self)
            self.fm = FingerManager(self)
            self.nm = NewsManager(self)
            self.om = OfferManager(self)
            self.alm = AutoLogOutManager(self)
        self.com = ConsoleManager(self)
        self.bm.start()
        self.players.start()
        self.games.start()
        self.seeks.start()
        self.challenges.start()

        # This block may useful if one wants to create
        # unit test lines from real life fics output
        if False:
            self.client.run_command("set seek 0")
            self.client.run_command("set shout 0")
            self.client.run_command("set cshout 0")
            self.client.run_command("iset seekinfo 0")
            self.client.run_command("iset seekremove 0")
            self.client.run_command("iset showownseek 0")
            self.client.run_command("iset allresults 0")
            self.client.run_command("iset pin 0")
            self.client.run_command("set open 0")
            self.client.run_command("set gin 0")
            self.client.run_command("set availinfo 0")

    def start_helper_manager(self, set_user_vars):
        # if guest accounts disabled we will handle players in the main connection
        if self.FatICS or self.USCN or self.ICC:
            self.client.run_command("set pin 1")
        else:
            self.client.run_command("iset allresults 1")
            # ivar pin: http://www.freechess.org/Help/HelpFiles/new_features.html
            self.client.run_command("iset pin 1")

        self.set_user_vars = set_user_vars
        if self.set_user_vars:
            self.client.run_command("set open 1")
            self.client.run_command("set gin 1")
            self.client.run_command("set availinfo 1")
        if self.ICC:
            self.hm = ICCHelperManager(self, self)
        else:
            self.hm = HelperManager(self, self)

        # disable setting iveriables from console
        self.client.run_command("iset lock 1")


class FICSHelperConnection(FICSConnection):
    def __init__(self, main_conn, host, ports, timeseal, username="guest", password=""):
        FICSConnection.__init__(self, host, ports, timeseal, username, password)
        self.main_conn = main_conn

    def _start_managers(self, lines):
        # The helper just wants only player and game notifications
        # set open 1 is a requirement for availinfo notifications
        self.client.run_command("set open 1")
        self.client.run_command("set seek 0")
        self.client.run_command("set shout 0")
        self.client.run_command("set cshout 0")
        self.client.run_command("set tell 0")
        self.client.run_command("set chanoff 1")
        self.client.run_command("set gin 1")
        self.client.run_command("set availinfo 1")
        if self.FatICS or self.USCN or self.ICC:
            self.client.run_command("set pin 1")
        else:
            self.client.run_command("iset allresults 1")
            # ivar pin: http://www.freechess.org/Help/HelpFiles/new_features.html
            self.client.run_command("iset pin 1")
        self.hm = HelperManager(self, self.main_conn)
